


def create_minimization_data(function, known_points, known_results, unknown_points, fixed_parameters):
	X = DataFrame.from_records(known_points + unknown_points)
	y = known_results + [function(**point, **fixed_parameters) for point in unknown_points]
	return X, y


def build_gaussian():
	return gaussian_process.GaussianProcessRegressor()


def sort_points(function, known_points, known_results, unknown_points, fixed_parameters, num_runs=1, model=None):
	known_points = known_points or []
	known_results = known_results or []

	num_runs = min(num_runs, len(unknown_points))
	if num_runs < 1:
		raise ValueError('no unknown point')

	new_unknown_points = unknown_points[num_runs:]

	new_known_points = unknown_points[:num_runs]
	new_known_results = [function(**point, **fixed_parameters) for point in new_known_points]
	model = model or gaussian_process.GaussianProcessRegressor()
	X_training = DataFrame.from_records(known_points + new_known_points)
	y_training = known_results + new_known_results
	model.fit(X_training, y_training)

	if len(new_unknown_points) > 0:
		X_test = DataFrame.from_records(new_unknown_points)
		try:
			y_test = model.predict(X_test)
		except Exception as e:
			display(X_test)
			raise e

		new_unknown_points = [x for _,x in sorted(zip(y_test, new_unknown_points), key=lambda pair: pair[0])]

	return {'known_points': new_known_points, 'known_results': new_known_results, 'unknown_points': new_unknown_points}



def minimize(
		function, model=None, parameter_ranges=None, parameter_values=None, fixed_parameters=None,
		num_starting_points=5, num_points_per_step=5, num_iterations=100,
		num_rounding_digits=6, seed=None
):
	"""
	:param callable function: function to be minimized
	:param dict[str, tuple] or NoneType parameter_ranges: parameters whose values should be chosen from a range
	:param dict[str, list] or NoneType parameter_values: parameters whose values should be chosen from a list
	:param dict[str, ] or NoneType fixed_parameters: parameters whose values are fixed
	:rtype: dict[str, ]
	"""
	parameter_ranges = parameter_ranges or {}
	parameter_values = parameter_values or {}
	fixed_parameters = fixed_parameters or {}

	points = get_random_points(
		parameter_ranges=parameter_ranges, parameter_values=parameter_values, num_random_points=num_iterations,
		num_rounding_digits=num_rounding_digits, seed=seed, max_tries=10
	)

	known = points[:num_starting_points]
	unknown = points[num_starting_points:]

	if


















from scipy import optimize
import numpy as np
import warnings


def bound_it(x, lower, upper):
	if lower is not None and upper is not None and lower > upper:
		lower, upper = upper, lower

	if lower is not None:
		x = max(x, lower)
	if upper is not None:
		x = min(x, upper)
	return x


class MinimizationResult:
	def __init__(self, result, arg_names, bounds=None, integers=None):
		self._result = result
		self._arg_names = arg_names
		self._bounds = bounds
		self._integers = integers

	@property
	def args(self):
		"""
		:rtype: dict
		"""
		if len(self._arg_names) == 1 and self._result.x.size == 1:
			result = {self._arg_names[0]: self._result.x.flat[0]}
		else:
			try:
				result = {key: value for key, value in zip(self._arg_names, self._result.x.flat)}
			except TypeError:

				display(self._arg_names, self._result.x)
				raise

		if self._integers:
			for key in self._integers:
				result[key] = int(round(result[key]))

		if self._bounds:
			for key in self._bounds:
				value = result[key]
				lower, upper = self._bounds[key]
				result[key] = bound_it(x=value, lower=lower, upper=upper)

		return result

	def __repr__(self):
		return repr(self.args)

	def __str__(self):
		return self.__repr__()


def penalize(x, upper, lower):
	"""
	produces a penalty relative to how much x is above the upper or below lower bound. maximum value is 2
	:param x:
	:param upper:
	:param lower:
	:return:
	"""
	if lower is not None and upper is not None and lower > upper:
		lower, upper = upper, lower
	u = 0
	l = 0
	if upper is not None and x > upper:
		u += 1 - (1 / ((x - upper)**2 + 1))
	if lower is not None and x < lower:
		l += 1 - (1 / ((x - lower)**2 + 1))
	return u + l


def minimize(
		function, args_to_optimize, initial_args, bounds=None, integers=None,
		constraint_penalty=1e3, constraints=None, method='Powell', max_iterations=None, max_evaluations=None,
		**kwargs
):
	"""
	a wrapper for the scipy minimize function that is easier to use because it uses dictionaries rather than lists
	:param callable				function: a function to be minimized
	:param list[str] 			args_to_optimize: list of function arguments to be optimized
	:param dict[str] 			initial_args: dictionary of initial values of the arguments
	:param dict 				kwargs: extra arguments for the scipy minimize method
	:param dict[str, tuple] 	bounds: dictionary of boundaries of the arguments, None means infinity
	:param list[str] 			integers: list of arguments that should be kept as integers
	:rtype: MinimizationResult
	"""
	constant_args = {key: value for key, value in initial_args.items() if key not in args_to_optimize}

	def new_function(x):
		new_kwargs = {key: value for key, value in zip(args_to_optimize, x)}
		new_kwargs.update(constant_args)
		penalty = 0

		if bounds is not None:
			for key in bounds:
				value = new_kwargs[key]
				lower, upper = bounds[key]
				new_kwargs[key] = bound_it(x=value, lower=lower, upper=upper)
				penalty += penalize(x=value, upper=upper, lower=lower)

		if integers is not None:
			for key in integers:
				try:
					new_kwargs[key] = int(round(new_kwargs[key]))
				except ValueError:
					pass

		return function(**new_kwargs) + penalty * constraint_penalty

	if integers is not None:
		integer_constraints = {
			'type': 'eq',
			'fun': lambda x: max([
				x[i] - int(x[i]) if key in integers else 0
				for i, key in enumerate(args_to_optimize)
			])
		}
		if constraints is None:
			constraints = (integer_constraints,)
		else:
			constraints = tuple(list(constraints) + [integer_constraints])

	x0 = np.array([initial_args[key] for key in args_to_optimize])

	fprime = lambda x: optimize.approx_fprime(x, new_function, 0.01)
	if max_iterations is not None or max_evaluations is not None:
		options = {}
		if max_iterations is not None:
			options['maxiter'] = max_iterations
		if max_evaluations is not None:
			options['maxfev'] = max_evaluations
	else:
		options = None

	try:
		with warnings.catch_warnings():
			warnings.simplefilter('ignore')
			result = optimize.minimize(
				fun=new_function, x0=x0, constraints=constraints, jac=fprime, method=method,
				options=options,
				**kwargs
			)
	except:
		with warnings.catch_warnings():
			warnings.simplefilter('ignore')
			result = optimize.minimize(
				fun=new_function, x0=x0, constraints=constraints, jac=fprime, method='Powell',
				options=options,
				**kwargs
			)

	return MinimizationResult(result=result, arg_names=args_to_optimize, bounds=bounds, integers=integers)



from pandas import DataFrame, Series
from copy import deepcopy


class PredictableData:
	def __init__(self, model, X):
		"""
		:param model:
		:type X: DataFrame
		"""
		self._model = model
		self._model_is_fitted = False
		self._X = X

	def fit(self, y=None, **kwargs):
		"""
		:type y: Series or list
		:type kwargs: dict
		:rtype: PredictableData
		"""
		if y is None:
			self._model.fit(self._X, **kwargs)
		else:
			self._model.fit(self._X, y, **kwargs)
		return self

	def predict(self, **kwargs):
		"""
		:type kwargs: dict
		:rtype: Series or list
		"""
		if hasattr(self._model, 'predict'):
			return self._model.predict(self._X, **kwargs)
		elif hasattr(self._model, 'pred'):
			return self._model.pred(self._X, **kwargs)
		else:
			raise AttributeError(f'{self._model} does not have any of predict or pred attributes!')

	@property
	def X(self):
		"""
		:rtype: DataFrame
		"""
		return self._X


class Predictor:
	def __init__(self, model, transformer=None, model_is_fitted=False, transformer_is_fitted=False, x_columns=None):
		self._transformer = transformer
		self._transformer_is_fitted = transformer_is_fitted or transformer is None
		self._unfitted_transformer = None
		self._model = model
		self._model_is_fitted = model_is_fitted
		self._unfitted_model = None
		self._x_columns = x_columns

	_STATE_ATTRIBUTES = ['_transformer', '_transformer_is_fitted', '_unfitted_transformer', '_model', '_model_is_fitted', '_unfitted_model', '_x_columns']

	def __getstate__(self):
		return {key: getattr(self, key) for key in self._STATE_ATTRIBUTES}

	def __setstate__(self, state):
		for key, value in state.items():
			setattr(self, key, value)

	@property
	def original_transformer(self):
		return self._unfitted_transformer or self._transformer

	@property
	def original_model(self):
		if self._unfitted_model is None:
			return self._model
		else:
			return self._unfitted_model

	def fit_transformer(self, X, **kwargs):
		"""
		:type X: DataFrame
		:type kwargs: dict
		:rtype: Predictor
		"""
		if self._transformer is not None:
			self._unfitted_transformer = deepcopy(self._transformer)
			self._transformer.fit(X, **kwargs)
		self._transformer_is_fitted = True
		self._x_columns = list(X.columns)
		return self

	def fit_model(self, X, y=None, **kwargs):
		"""
		:type X: DataFrame
		:type y: NoneType or Series or list
		:type kwargs: dict
		:rtype: Predictor
		"""
		self._unfitted_model = deepcopy(self._model)
		self.transform(X=X).fit(y=y, **kwargs)
		self._model_is_fitted = True
		if self._transformer is None:
			self._x_columns = list(X.columns)
		return self

	def fit(self, X, y=None, **kwargs):
		if self._transformer_is_fitted:
			self.fit_model(X=X, y=y, **kwargs)
		else:
			self.fit_transformer(X=X).fit_model(X=X, y=y, **kwargs)

	def transform(self, X, **kwargs):
		"""
		:type X: DataFrame
		:rtype: PredictableData
		"""
		if not self._transformer_is_fitted:
			raise RuntimeError('transformer is not fitted!')
		if self._x_columns:
			X = X[self._x_columns]

		if self._transformer is None:
			transformed = X
		elif hasattr(self._transformer, 'transform'):
			transformed = self._transformer.transform(X, **kwargs)
		elif hasattr(self._transformer, 'predict'):
			transformed = self._transformer.predict(X, **kwargs)
		elif hasattr(self._transformer, 'pred'):
			transformed = self._transformer.pred(X, **kwargs)
		else:
			raise AttributeError(f'{self._transformer} does not have any of transform, predict, or pred attributes!')

		return PredictableData(model=self._model, X=transformed)

	def predict(self, X, **kwargs):
		"""
		:type X: DataFrame
		:rtype: Series or list
		"""
		return self.transform(X=X).predict(**kwargs)



from .get_coefficients import get_coefficients
from .get_feature_importances import get_feature_importances
from .InfluenceSimulator import InfluenceSimulator
from .get_model_influence import get_model_influence


from pandas import DataFrame
from copy import deepcopy
from functools import wraps


def get_values_from_range(start, end, num_digits):
	result = []
	x = start
	step = 1/(10**num_digits)
	while x <= end:
		result.append(round(x, num_digits))
		x += step
	return result


def create_mimic(function, model, known_points, known_results, unknown_points, fixed_parameters=None, return_results=False):
	"""
	:type function: callable 
	:rtype: callable or tuple(callable, list)
	"""
	fixed_parameters = fixed_parameters or {}
	results_of_unknown = [function(**unknown_point, **fixed_parameters) for unknown_point in unknown_points]

	X = DataFrame.from_records(list(known_points) + list(unknown_points))
	y = list(known_results) + results_of_unknown

	new_model = deepcopy(model)
	new_model.fit(X, y)

	x_columns = X.columns

	@wraps(function)
	def mimic(**kwargs):
		for key in set(kwargs.keys()).union(set(x_columns)):
			if key not in x_columns:
				raise KeyError(f'{key} is does not exist in the original function')
			elif key not in kwargs.keys():
				raise KeyError(f'{key} is missing from the arguments')

		X = DataFrame.from_records([kwargs])
		return new_model.predict(X)[0]

	if return_results:
		return mimic, results_of_unknown
	else:
		return mimic




from slytherin.collections import create_grid
import random
from pandas import DataFrame
from sklearn import gaussian_process


def get_values_from_range(start, end, num_digits):
	result = []
	x = start
	step = 1/(10**num_digits)
	while x <= end:
		result.append(round(x, num_digits))
		x += step
	return result


def get_all_points(parameter_ranges, parameter_values, num_rounding_digits):
	parameter_range_values = {
		key: get_values_from_range(start=range_tuple[0], end=range_tuple[1], num_digits=num_rounding_digits)
		for key, range_tuple in parameter_ranges.items()
	}
	parameters = create_grid({**parameter_range_values, **parameter_values})
	return parameters


def get_minimization_points(
		parameter_ranges, parameter_values, num_random_points, num_rounding_digits, seed, max_tries=10
):
	all_points = []
	tries = 0
	while len(all_points) < num_random_points and tries <= max_tries:

		all_points = get_all_points(
			parameter_ranges=parameter_ranges, parameter_values=parameter_values,
			num_rounding_digits=num_rounding_digits + tries
		)
		tries += 1
		if len(parameter_ranges) == 0:
			break

	if seed is not None:
		random.seed(seed)
	random.shuffle(all_points)
	return all_points[:num_random_points]



import statsmodels.formula.api as smf
from .exceptions import BrokenModel
from .Regression import Regression


class OLS(Regression):

	def __init__(self, data, formula, significance_level=0.05, model_builder=None, family=None, groups=None, parent=None):
		super().__init__(data=data, formula=formula, model_builder=smf.ols, significance_level=significance_level, parent=parent)

	@property
	def fit(self):
		"""
		:rtype: OLS
		"""
		if self._fit is None:
			try:
				self._fit = self.model.fit(disp=False)
			except Exception as e:
				self._fit = BrokenModel(exception=e, regression=self)

		return self._fit

	@property
	def name(self):
		"""
		:rtype: str
		"""
		return 'ols'

	@property
	def coefficient_table(self):
		"""
		:rtype: DataFrame
		"""
		return super().coefficient_table.rename(columns={'P>|t|': 'p'})

	@property
	def r_squared(self):
		"""
		:rtype: float
		"""
		return self.fit.r_squared

	@property
	def adjusted_r_squared(self):
		"""
		:rtype: float
		"""
		return self.fit.adjusted_r_squared




from .minimize import minimize
from .create_mimic import create_mimic


def fast_minimize(
		function, args_to_optimize, model, random_points, num_starting_points, fixed_parameters=None, num_repeats=1,
		bounds=None, integers=None,
		constraint_penalty=1e3, constraints=None, method='Powell', max_iterations=None, max_evaluations=None,
		**kwargs
):
	known_points = []
	known_results = []
	unknown_points = random_points[:num_starting_points]
	remaining_random_points = random_points[num_starting_points:]
	fixed_parameters = fixed_parameters or {}
	minimization = None

	for i in range(num_repeats):
		mimic, unknown_results = create_mimic(
			function=function,
			model=model,
			known_points=known_points,
			known_results=known_results,
			unknown_points=unknown_points,
			fixed_parameters=fixed_parameters,
			return_results=True
		)
		known_points += unknown_points
		known_results += unknown_results

		the_zip = list(sorted(zip(known_points, known_results), key=lambda pair: pair[1]))
		best_point = the_zip[0][0]
		#print('best_point:', best_point, 'result:', the_zip[0][1])
		minimization = minimize(
			function=mimic, args_to_optimize=args_to_optimize, initial_args=best_point,
			bounds=bounds, integers=integers,
			constraints=constraints, constraint_penalty=constraint_penalty, method=method,
			max_iterations=max_iterations, max_evaluations=max_evaluations, **kwargs
		)
		#print('minimization:', minimization)
		minimization_point = minimization.args
		minimization_result = function(**minimization_point, **fixed_parameters)
		known_points.append(minimization_point)
		known_results.append(minimization_result)
		if len(remaining_random_points) > 0:
			random_point = remaining_random_points.pop()
			random_result = function(**random_point, **fixed_parameters)

			#if minimization_result < random_result:
			#	print('minimization is smaller')
			#else:
			#	print('random is smaller')
			known_points.append(random_point)
			known_results.append(random_result)
		else:
			pass
			#print('ran out of random points')
		unknown_points = []

	return minimization#, known_points, known_results
